#===============================================================================
#  mean.py
#===============================================================================

"""Calculate average methylation level across a genome"""




# Imports ======================================================================

from pybedtools import BedTool

import pandas as pd
import seaborn as sns




# Constants ====================================================================

COLOR_PALETTE = sns.color_palette().as_hex()




# Functions ====================================================================

def generate_data(bedtools, groups, chromosomes=None):
    """"Generator giving data from bedmethyl files

    Parameters
    ----------
    bedtools
        iterable of BedTools
    groups
        iterable giving group id for each input bedtool
    chromosomes
        iterable of chromosomes to include, or None to include all chromosomes

    Yields
    ------
    tuple
        ((chrom, methylation_level), group)
    """
    if chromosomes is not None:
        chromosomes = set(chromosomes)
    for bt, g in zip(bedtools, groups):
        for i in bt:
            n_fields = len(i.fields)
            if n_fields == 10:
                chrom, _, _, _, _, _, _, _, _, meth_col, group = i.fields + [g]
                meth = meth_col.split()[1]
            elif n_fields == 9:
                chrom, _, _, meth, _, _, _, _, _, group = i.fields + [g]
            else:
                raise RuntimeError('Invalid methylation BED file')
            if (chromosomes is None) or (chrom in chromosomes):
                yield chrom, float(meth), group


def calculate_mean(bedmethyl, plot=None, chromosomes=None, groups=None,
              title='Methylation', legend_title='Group', width=8,
              color_palette=COLOR_PALETTE):
    """Calculate mean methylation level for each chromosome

    Parameters
    ----------
    bedmethyl
        iterable of paths to bedmethyl formatted files containing methylation results
    plot
        if given, path to output plot
    chromosomes
        iterable of chromosomes to include, or None to include all chromosomes
    groups
        iterable of group ids for bedmethyl files
    title : str
        plot title
    legend_title : str
        plot legend title
    width : float
        plot width
    """


    if not groups:
        groups = list(range(len(bedmethyl)))
    methyl = tuple(BedTool(bed) for bed in bedmethyl)
    plotting_data = pd.DataFrame(
        generate_data(methyl, groups, chromosomes),
        columns=('Chromosome', 'Methylation level (%)', legend_title))
    chromosome_means = plotting_data.groupby(by=['Chromosome', legend_title]).mean()
    print('Chromosome', legend_title, 'Methylation level (%)', sep='\t')
    for (chrom, group), row in chromosome_means.iterrows():
        print(chrom, group, row[0], sep='\t')

    if plot:
        ax = sns.barplot(x='Chromosome', y='Methylation level (%)',
                         hue=legend_title, data=plotting_data, errorbar=None,
                         palette=color_palette)
        ax.set_title(title)

        ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left', borderaxespad=0,
                  title=legend_title)
        ax.set_xticks(ax.get_xticks())
        xtl = ax.get_xticklabels()
        ax.set_xticklabels(xtl, rotation=30, horizontalalignment='right')
        fig = ax.get_figure()
        fig.set_figheight(3)
        fig.set_figwidth(width)
        fig.tight_layout()
        fig.savefig(plot)
        fig.clf()


def calculate_total_mean(bedmethyl, plot=None, groups=None,
              title='Methylation', legend_title='Group', width=3,
              color_palette=COLOR_PALETTE):
    """Calculate total genomic mean methylation level

    Parameters
    ----------
    bedmethyl
        iterable of paths to bedmethyl formatted files containing methylation results
    plot
        if given, path to output plot
    groups
        iterable of group ids for bedmethyl files
    title : str
        plot title
    legend_title : str
        plot legend title
    width : float
        plot width
    """


    if not groups:
        groups = list(range(len(bedmethyl)))
    methyl = tuple(BedTool(bed) for bed in bedmethyl)
    plotting_data = pd.DataFrame(
        generate_data(methyl, groups, chromosomes=None),
        columns=('Chromosome', 'Methylation level (%)', legend_title))
    total_means = plotting_data.loc[:, ['Methylation level (%)', legend_title]].groupby(
        by=legend_title).mean()
    print(legend_title, 'Methylation level (%)', sep='\t')
    for group, row in total_means.iterrows():
        print(group, row[0], sep='\t')

    if plot:
        ax = sns.barplot(x=legend_title, y='Methylation level (%)',
                         hue=legend_title, data=plotting_data, ci=None,
                         palette=color_palette, dodge=False)
        ax.set_title(title)
        ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left', borderaxespad=0,
                  title=legend_title)
        fig = ax.get_figure()
        fig.set_figheight(2)
        fig.set_figwidth(width)
        fig.tight_layout()
        fig.savefig(plot)
        fig.clf()



#===============================================================================
# loreme.py
#===============================================================================

"""Process PacBio or Nanopore methylation data"""




# Imports ======================================================================

import argparse
import os.path
import platform
import shutil
from math import floor
from loreme.env import (PBCPG_DIR, PBCPG_MODEL, EXAMPLE_DATA_DIR, DORADO_DIR,
                       DORADO_MODEL_DIR, DORADO_PLATFORM, MODKIT_DIR)
from loreme.version import __version__
from loreme.download import (download_pbcpg, download_pb_example,
                             download_dorado, download_modkit,
                             download_ont_example)
from loreme.check_tags import check_tags
from loreme.check_gpu_availability import check_gpu_availability
from loreme.pbcpg import pbcpg_align_bams, pbcpg_pileup
from loreme.dorado import CPU_COUNT, fast5_to_pod5, dorado_basecall, dorado_align
from loreme.modkit import modkit_pileup
from loreme.mean import calculate_mean, calculate_total_mean
from loreme.gene_body_methylation import gene_body_methylation
from loreme.promoter_methylation import promoter_methylation
from loreme.plot import plot, COLOR_PALETTE
from loreme.plot_genes import plot_genes
from loreme.plot_repeats import plot_repeats
from loreme.merge import merge_bedmethyl
from loreme.intersect import intersect_bedmethyl
from loreme.export_bedgraph import export_bedgraph




# Functions ====================================================================

def error_exit(msg):
    """Exit with an error

    Parameters
    ----------
    msg : str
        String describing the error
    """

    raise RuntimeError(msg)


def pbcpg_validate_args_pre_alignment(args):
    """Validate arguments before the alignment step

    Parameters
    ----------
    args
        argparse.Namespace containing the arguments
    """

    def check_required_file(file, label):
        if not os.path.isfile(file):
            error_exit(f"Can't find {label} file '{file}'")
    for bam in args.bam:
        check_required_file(bam, "input bam")
    check_required_file(args.fasta, "reference fasta")
    if not os.path.isfile(args.model):
        error_exit("{} is not a valid model file!".format(args.model))


def pbcpg_validate_args_post_alignment(args):
    """Validate arguments after the alignment step

    Parameters
    ----------
    args
        argparse.Namespace containing the arguments
    """

    def is_bam_index_found(bam_file):
        bam_index_extensions = (".bai", ".csi")
        for ext in bam_index_extensions:
            bam_index_file=bam_file+ext
            if os.path.isfile(bam_index_file):
                return True
        return False
    if not is_bam_index_found(args.bam):
        error_exit(f"Can't find index for bam file '{args.bam}'")


def _download_pbcpg(args):
    download_pbcpg(directory=args.directory)


def _download_pb_example(args):
    download_pb_example(directory=args.directory, n_samples=args.n_samples)


def _download_ont_example(args):
    download_ont_example(directory=args.directory)


def _download_dorado(args):
    download_dorado(args.platform, directory=args.directory,
                    model_dir=args.model_dir)


def _download_modkit(args):
    download_modkit(directory=args.directory)


def _check_tags(args):
    if check_tags(args.bam, n_reads=args.n_reads):
        print('MM/ML tags found')
    else:
        raise RuntimeError('MM/ML tags not found in BAM file')


def _check_gpu_availability(args):
    if check_gpu_availability(args.devices, gpu_load=args.gpu_load,
                           gpu_mem=args.gpu_mem):
        print('GPU is available')


def _pbcpg_align(args):
    pbcpg_align_bams(args.fasta, args.bam, args.aligned_bam,
        threads=max(1, args.threads-1),
        memory_mb=max(floor(args.memory/args.threads), 1))


def _pbcpg_pileup(args):
    pbcpg_pileup(args.bam, args.output_prefix, hap_tag=args.hap_tag,
                  min_coverage=args.min_coverage,min_mapq=args.min_mapq,
                  model=args.model, threads=args.threads)


def run_pbcpg(args):
    pbcpg_validate_args_pre_alignment(args)
    for bam in args.bam:
        if not check_tags(bam, n_reads=args.n_reads):
            raise RuntimeError('MM/ML tags not found in BAM file')
    aligned_bam = f'{args.output_prefix}.pbmm2.bam'
    pbcpg_align_bams(args.fasta, args.bam, aligned_bam,
        threads=max(1, args.threads-1),
        memory_mb = max(floor(args.memory/args.threads), 1))
    args.bam = aligned_bam
    pbcpg_validate_args_post_alignment(args)
    _pbcpg_pileup(args)


def _fast5_to_pod5(args):
    fast5_to_pod5(*args.fast5, output_pod5=args.output, threads=args.threads)


def _dorado_align(args):
    dorado_align(args.index, args.reads, args.output, threads=args.threads,
                 mem_per_thread_mb=int(args.memory_mb/args.threads))


def _modkit_pileup(args):
    modkit_pileup(args.fasta, args.input, args.output, threads=args.threads)


def run_dorado_basecall(args):
    check_gpu_availability([0])
    if args.convert:
        pod5_dir = f'{args.output[:-4]}_pod5'
        pod5_base = f'{os.path.basename(args.output)[:-4]}.pod5'
        os.mkdir(pod5_dir)
        fast5_to_pod5(*args.reads,
                      output_pod5=os.path.join(pod5_dir, pod5_base),
                      threads=args.threads)
        dorado_basecall(pod5_dir, args.output, speed=args.speed,
                        accuracy=args.accuracy, no_mod=args.no_mod)
    elif len(args.reads) == 1 and os.path.isdir(args.reads[0]):
        dorado_basecall(args.reads[0], args.output, speed=args.speed,
               accuracy=args.accuracy, no_mod=args.no_mod)
    else:
        raise RuntimeError('For running without --convert, supply only one input FAST5 or POD5 directory')



def _calculate_mean(args):
    if args.total:
        calculate_total_mean(args.bedmethyl, plot=args.plot,
                             groups=args.groups, title=args.title,
                             legend_title=args.legend_title, width=args.width,
                             color_palette=args.color_palette)
    else:
        calculate_mean(args.bedmethyl, plot=args.plot,
                       chromosomes=args.chromosomes, groups=args.groups,
                       title=args.title, legend_title=args.legend_title,
                       width=args.width, color_palette=args.color_palette)


def _promoter_methylation(args):
    promoter_methylation(args.features, args.bedmethyl,
        upstream_flank=args.upstream_flank,
        downstream_flank=args.downstream_flank, chromosomes=args.chromosomes,
        cytosines=args.cytosines, coverage=args.coverage,
        min_coverage=args.min_coverage, bins=args.bins, levels=args.levels,
        hist=args.hist, hist_log=args.hist_log, palette=args.palette)


def _gene_body_methylation(args):
    gene_body_methylation(args.features, args.bedmethyl,
        upstream_flank=args.upstream_flank,
        downstream_flank=args.downstream_flank, chromosomes=args.chromosomes,
        cytosines=args.cytosines, coverage=args.coverage,
        min_coverage=args.min_coverage, bins=args.bins, levels=args.levels,
        hist=args.hist, hist_log=args.hist_log, palette=args.palette)

def _plot(args):
    plot(args.bedmethyl, args.output, reference=args.reference,
         chromosomes=args.chromosomes, groups=args.groups, title=args.title,
         legend=args.legend, legend_title=args.legend_title,
         bin_size=args.bin_size, width=args.width,
         color_palette=args.color_palette, alpha=args.alpha,
         x_label=args.x_label)


def _plot_genes(args):
    plot_genes(args.features, args.bedmethyl, args.output, groups=args.groups,
               flank=args.flank, smooth=args.smooth, title=args.title,
               confidence_interval=args.confidence_interval,
               gene_bins=args.gene_bins, gene_levels=args.gene_levels,
               legend=args.legend, legend_title=args.legend_title,
               palette=args.palette, width=args.width, alpha=args.alpha)


def _plot_repeats(args):
    plot_repeats(args.features, args.bedmethyl, args.output, type=args.type,
                 groups=args.groups, flank=args.flank, smooth=args.smooth,
                 title=args.title, confidence_interval=args.confidence_interval,
                 legend=args.legend, legend_title=args.legend_title,
                 palette=args.palette, width=args.width, alpha=args.alpha)

def _merge(args):
    merge_bedmethyl(args.bedmethyl, chromosomes=args.chromosomes)


def _intersect(args):
    intersect_bedmethyl(args.bedmethyl, chromosomes=args.chromosomes)


def _export_bedgraph(args):
    export_bedgraph(args.bedmethyl, chromosomes=args.chromosomes,
                    coverage=args.coverage)

def clean(args):
    for root in args.dirs:
        for file in ('pb-CpG-tools-v2.3.1-x86_64-unknown-linux-gnu.tar.gz',
                    #  'dorado-0.2.4-linux-x64.tar.gz',
                     'dorado-0.3.2-linux-x64.tar.gz',
                     'modkit_v0.1.8_centos7_x86_64.tar.gz',
                     'dorado-config.json'):
            if os.path.isfile(os.path.join(root, file)):
                os.remove(os.path.join(root, file))
        for directory in ('pb-CpG-tools-v2.3.1-x86_64-unknown-linux-gnu/',
                        #   f'dorado-0.2.4-{DORADO_PLATFORM}/',
                          f'dorado-0.3.2-{DORADO_PLATFORM}/',
                          'dna_r10.4.1_e8.2_260bps_fast@v4.1.0/',
                          'dna_r10.4.1_e8.2_260bps_hac@v4.1.0/',
                          'dna_r10.4.1_e8.2_260bps_sup@v4.1.0/',
                          'dna_r10.4.1_e8.2_400bps_fast@v4.1.0/',
                          'dna_r10.4.1_e8.2_400bps_hac@v4.1.0/',
                          'dna_r10.4.1_e8.2_400bps_sup@v4.1.0/',
                          'dna_r10.4.1_e8.2_260bps_fast@v4.1.0_5mCG_5hmCG@v2/',
                          'dna_r10.4.1_e8.2_260bps_hac@v4.1.0_5mCG_5hmCG@v2/',
                          'dna_r10.4.1_e8.2_260bps_sup@v4.1.0_5mCG_5hmCG@v2/',
                          'dna_r10.4.1_e8.2_400bps_fast@v4.1.0_5mCG_5hmCG@v2/',
                          'dna_r10.4.1_e8.2_400bps_hac@v4.1.0_5mCG_5hmCG@v2/',
                          'dna_r10.4.1_e8.2_400bps_sup@v4.1.0_5mCG_5hmCG@v2/',
                          'dist'):
            if os.path.exists(os.path.join(root, directory)):
                shutil.rmtree(os.path.join(root, directory))


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="""Analysis of DNA methylation signals from `Pacific Biosciences <https://www.pacb.com/technology/hifi-sequencing/>`_
or `Oxford Nanopore <https://nanoporetech.com/applications/dna-nanopore-sequencing>`_
long read sequencing data.""")
    parser.add_argument('--version', action='version',
                    version='%(prog)s {version}'.format(version=__version__))
    subparsers = parser.add_subparsers()

    parser_download_pbcpg = subparsers.add_parser('download-pbcpg',
                                                  help='download pb-CpG-tools')
    parser_download_pbcpg.set_defaults(func=_download_pbcpg)
    parser_download_pbcpg.add_argument(
        '-d', '--directory', metavar='<directory/>', default=PBCPG_DIR,
        help=f'destination directory for pb-CpG-tools (default: {PBCPG_DIR})')

    parser_download_dorado = subparsers.add_parser('download-dorado',
                                                  help='download dorado')
    parser_download_dorado.set_defaults(func=_download_dorado)
    parser_download_dorado.add_argument(
        'platform', choices=('linux-x64', 'linux-arm64', 'osx-arm64', 'win64'),
        help=f'Select your platform. If you\'re unsure, the following information about your system and machine may help you choose. system: {platform.system()}, machine: {platform.machine()}')
    parser_download_dorado.add_argument(
        '-d', '--directory', metavar='<directory/>', default=DORADO_DIR,
        help=f'destination directory for dorado (default: {DORADO_DIR})')
    parser_download_dorado.add_argument(
        '-m', '--model-dir', metavar='<directory/>', default=DORADO_MODEL_DIR,
        help=f'destination directory for dorado models (default: {DORADO_MODEL_DIR})')

    parser_download_modkit = subparsers.add_parser('download-modkit',
                                                  help='download modkit')
    parser_download_modkit.set_defaults(func=_download_modkit)
    parser_download_modkit.add_argument(
        '-d', '--directory', metavar='<directory/>', default=MODKIT_DIR,
        help=f'destination directory for pb-CpG-tools (default: {MODKIT_DIR})')

    parser_download_pb_example = subparsers.add_parser('download-pb-example',
        help='download an example dataset for pbcpg')
    parser_download_pb_example.set_defaults(func=_download_pb_example)
    parser_download_pb_example.add_argument( '-d', '--directory', metavar='<directory/>',
        default=EXAMPLE_DATA_DIR, help='destination directory for example data')
    parser_download_pb_example.add_argument('-n', '--n-samples', metavar='<int>',
        type=int, default=1, help='number of samples to download, max 4 (default: 1)')
    
    parser_download_ont_example = subparsers.add_parser('download-ont-example',
        help='download an example dataset for dorado')
    parser_download_ont_example.set_defaults(func=_download_ont_example)
    parser_download_ont_example.add_argument( '-d', '--directory', metavar='<directory/>',
        default=EXAMPLE_DATA_DIR, help='destination directory for example data')

    parser_check_tags = subparsers.add_parser('check-tags', help='check BAM for MM/ML tags')
    parser_check_tags.set_defaults(func=_check_tags)
    parser_check_tags.add_argument("bam", metavar="<reads.bam>",
        help="BAM file to check.")
    parser_check_tags.add_argument('-n', '--n-reads', metavar="<int>",
        type=int, default=1000, help="Number of reads to check (default: %(default)d].")

    parser_check_gpu = subparsers.add_parser('check-gpu',
        help='check GPU availability')
    parser_check_gpu.set_defaults(func=_check_gpu_availability)
    parser_check_gpu.add_argument('--devices', metavar='<int>', type=int,
                                      nargs='+', default=[0],
                                      help='GPUs to check')
    parser_check_gpu.add_argument('--gpu-load', metavar='<float>', type=float,
                                     default=0.9,
                                     help='required available GPU load')
    parser_check_gpu.add_argument('--gpu-mem', metavar='<float>', type=float,
                                     default=0.9,
                                     help='required available GPU memory')

    parser_pbcpg = subparsers.add_parser('pbcpg', help='run full pbcpg pipeline')
    parser_pbcpg.set_defaults(func=run_pbcpg)
    io_args = parser_pbcpg.add_argument_group('io args')
    io_args.add_argument("bam", metavar="<unaligned.bam>", nargs='+',
                        help="Unaligned BAM file.")
    io_args.add_argument("fasta", metavar="<ref.fasta>",
                        help="The reference fasta file.")
    io_args.add_argument("output_prefix", metavar="<prefix>",
                        help="Label for output files, which results in [prefix].bam/bed/bw.")
    score_args = parser_pbcpg.add_argument_group('score args')
    score_args.add_argument("--model", metavar="</path/to/model.tflite>",
                        default=PBCPG_MODEL,
                        help=f"Full path to the directory containing the model to load. (default: {PBCPG_MODEL})")
    score_args.add_argument("-m", "--modsites", choices=["denovo", "reference"],
                        default="denovo",
                        help="Only output CG sites with a modification probability > 0 "
                             "(denovo), or output all CG sites based on the "
                             "supplied reference fasta (reference). (default: %(default)s)")
    score_args.add_argument("-c", "--min_coverage", metavar="<int>", default=4,
                        type=int,
                        help="Minimum coverage required for filtered outputs. (default: %(default)d)")
    score_args.add_argument("-q", "--min-mapq", metavar="<int>", default=0,
                        type=int,
                        help="Ignore alignments with MAPQ < N. (default: %(default)d)")
    score_args.add_argument("-a", "--hap-tag", metavar="<TAG>", default="HP",
                        help="The SAM tag containing haplotype information. (default: %(default)s)")
    score_args.add_argument("-s", "--chunksize", metavar="<int>", default=500_000,
                        type=int,
                        help="Break reference regions into chunks "
                             "of this size for parallel processing. (default: %(default)d)")
    score_args.add_argument('-n', '--n-reads', metavar="<int>",
                        type=int, default=1000,
                        help="Number of reads for MM/ML tag check. (default: %(default)d)")
    resource_args = parser_pbcpg.add_argument_group('resource args')
    resource_args.add_argument("-t", "--threads", metavar="<int>", default=1,
                        type=int,
                        help="Number of threads for parallel processing. (default: %(default)d)")
    resource_args.add_argument("--memory", metavar="<int>", default=4_000,
                        type=int,
                        help="Memory for read alignment and sorting in megabytes. (default: %(default)d)")

    parser_pbcpg_align = subparsers.add_parser('pbcpg-align', help='align BAM to reference')
    parser_pbcpg_align.set_defaults(func=_pbcpg_align)
    io_args = parser_pbcpg_align.add_argument_group('io args')
    io_args.add_argument("bam", metavar="<unaligned.bam>", nargs='+',
        help="Unaligned BAM file to read.")
    io_args.add_argument("fasta", metavar="<ref.fasta>",
        help="The reference fasta file.")
    io_args.add_argument("aligned_bam", metavar="<aligned.bam>",
        help="Aligned bam file to write.")
    resource_args = parser_pbcpg_align.add_argument_group('resource args')
    resource_args.add_argument("-t", "--threads", metavar="<int>", type=int,
        default=1,
        help="Number of threads for parallel processing. (default: %(default)d)")
    resource_args.add_argument("--memory", metavar="<int>", type=int,
        default=4_000,
        help="Memory for read alignment and sorting in megabytes. (default: %(default)d)")

    parser_pbcpg_pileup = subparsers.add_parser('pbcpg-pileup', help='pile up methylation calls from aligned PB reads')
    parser_pbcpg_pileup.set_defaults(func=pbcpg_pileup)
    io_args = parser_pbcpg_pileup.add_argument_group('io args')
    io_args.add_argument("bam", metavar="<aligned.bam>",
                        help="Aligned BAM file.")
    io_args.add_argument("fasta", metavar="<ref.fasta>",
                        help="The reference fasta file.")
    io_args.add_argument("output_prefix", metavar="<prefix>",
                        help="prefix for output files, which results in [prefix].bam/bed/bw.")
    score_args = parser_pbcpg_pileup.add_argument_group('score args')
    score_args.add_argument("--model", metavar="</path/to/model.tflite>",
                        default=PBCPG_MODEL,
                        help=f"Full path to the model (*.tflite files) to load. (default: {PBCPG_MODEL})")
    score_args.add_argument("--modsites", choices=["denovo", "reference"],
                        default="denovo",
                        help="Only output CG sites with a modification probability > 0 "
                             "(denovo), or output all CG sites based on the "
                             "supplied reference fasta (reference). (default: %(default)s)")
    score_args.add_argument("-c", "--min_coverage", metavar="<int>", default=4,
                        type=int,
                        help="Minimum coverage required for filtered outputs. (default: %(default)d)")
    score_args.add_argument("-q", "--min_mapq", metavar="<int>", default=0,
                        type=int,
                        help="Ignore alignments with MAPQ < N. (default: %(default)d)")
    score_args.add_argument("-a", "--hap_tag", metavar="<TAG>", default="HP",
                        help="The SAM tag containing haplotype information. (default: %(default)s)")
    score_args.add_argument("-s", "--chunksize", metavar="<int>", default=500_000,
                        type=int,
                        help="Break reference regions into chunks "
                             "of this size for parallel processing. (default: %(default)d)")
    resource_args = parser_pbcpg_pileup.add_argument_group('resource args')
    resource_args.add_argument("-t", "--threads", metavar="<int>", default=1,
                        type=int,
                        help="Number of threads for parallel processing. (default: %(default)d)")

    parser_dorado_convert = subparsers.add_parser('dorado-convert',
        help='convert FAST5 to POD5')
    parser_dorado_convert.set_defaults(func=_fast5_to_pod5)
    parser_dorado_convert.add_argument('fast5', metavar='<reads[.fast5]>', nargs='+',
                               help='input FAST5 files, or direcories containing FAST5')
    parser_dorado_convert.add_argument('output', metavar='<output.pod5>',
                               help='path to output POD5 file')
    parser_dorado_convert.add_argument("-t", "--threads", metavar="<int>",
        type=int, default=1, help="Number of threads. (default: %(default)d)")

    parser_dorado_basecall = subparsers.add_parser('dorado-basecall',
        help='run dorado basecaller')
    parser_dorado_basecall.set_defaults(func=run_dorado_basecall)
    parser_dorado_basecall.add_argument('reads', metavar='<reads>', nargs='+',
                               help='input FAST5 files, or single directory containing FAST5 or POD5 files')
    parser_dorado_basecall.add_argument('output', metavar='<output.sam>',
                               help='path to output SAM file')
    parser_dorado_basecall.add_argument('--convert', action='store_true',
                               help='convert FAST5 to POD5 before basecalling')
    parser_dorado_basecall.add_argument('--speed', type=int, choices=(260, 400),
                               default=400, help='pore speed (default: 400)')
    parser_dorado_basecall.add_argument('--accuracy', choices=('fast', 'hac', 'sup'),
                               default='fast', help='model accuracy (default: fast)')
    parser_dorado_basecall.add_argument('--no-mod', action='store_true',
                               help='turn off modified basecalling')
    # parser_dorado_basecall.add_argument('--frequency', choices=('4kHz', '5kHz'),
    #                            default='4kHz', help='(default: 4kHz)')
    # parser_dorado_basecall.add_argument('--modified-bases',
    #                            choices=('5mCG_5hmCG', '5mC', '6mA'),
    #                            default='5mCG_5hmCG',
    #                            help='Modified base model (default: 5mCG_5hmCG)')
    # parser_dorado_basecall.add_argument('--reference', metavar='<index>',
    #                            help='map to a reference index (fastq/fasta/mmi)')
    parser_dorado_basecall.add_argument("-t", "--threads", metavar="<int>",
        type=int, default=1,
        help="Number of cpu threads. (default: %(default)d)")

    parser_dorado_align = subparsers.add_parser('dorado-align',
        help='run dorado aligner')
    parser_dorado_align.set_defaults(func=_dorado_align)
    parser_dorado_align.add_argument('index', metavar='<index>',
                                     help='reference index for alignment')
    parser_dorado_align.add_argument('reads', metavar='<reads>',
                                     help='reads for alignment')
    parser_dorado_align.add_argument('output', metavar='<output.bam>',
                                     help='path to output BAM file')
    parser_dorado_align.add_argument('--threads', metavar='<int>', type=int,
                                     default=CPU_COUNT,
                                     help=f'number of sorting threads (default: {CPU_COUNT})')
    parser_dorado_align.add_argument('--memory-mb', metavar='<int>', type=int,
                                     default=CPU_COUNT*768,
                                     help=f'approximate sorting memory in MB (default: {CPU_COUNT*768})')
    
    parser_modkit_pileup = subparsers.add_parser('modkit-pileup',
        help='pile up aligned dorado methylation calls')
    parser_modkit_pileup.set_defaults(func=_modkit_pileup)
    parser_modkit_pileup.add_argument('fasta', metavar='<reference.fasta>',
        help='reference FASTA file')
    parser_modkit_pileup.add_argument('input', metavar='<input.bam>',
        help='input BAM file')
    parser_modkit_pileup.add_argument('output', metavar='<output.bed>',
        help='output BED file')
    parser_modkit_pileup.add_argument("-t", "--threads", metavar="<int>",
        type=int, default=1, help="Number of threads. (default: %(default)d)")

    parser_mean = subparsers.add_parser('mean',
        help='calculate average methylation across chromosomes or in total')
    parser_mean.set_defaults(func=_calculate_mean)
    parser_mean.add_argument('bedmethyl', metavar='<bedmethyl.bed>',
        nargs='+', help='bedMethyl file containing methylation data')
    parser_mean.add_argument('--plot', metavar='<output.{pdf,png,svg}>',
        help='path to output plot')
    domain_group = parser_mean.add_mutually_exclusive_group()
    domain_group.add_argument('--chromosomes', metavar='<X>', nargs='+',
        help='chromosomes to include')
    domain_group.add_argument('--total', action='store_true',
        help='calculate total genomic mean')
    parser_mean.add_argument('--groups', metavar='<"Group">', nargs='+',
        help='list of groups for provided bedmethyl files [0)')
    parser_mean.add_argument('--title', metavar='<"Plot title">',
        default='Methylation', help='set the title for the plot')
    parser_mean.add_argument('--legend-title', metavar='<Title>',
        default='Group', help='title of legend')
    parser_mean.add_argument('--width', metavar='<float>', type=float,
        default=8, help='set width of figure in inches')
    parser_mean.add_argument('--color-palette', metavar='<#color>', nargs='+',
        default=COLOR_PALETTE, help='color palette to use')

    parser_promoter = subparsers.add_parser('promoter',
        help='quantify promoter methylation')
    parser_promoter.set_defaults(func=_promoter_methylation)
    parser_promoter.add_argument('features', metavar='<features.gff3>',
        help='gff3 file of genomic features')
    parser_promoter.add_argument('bedmethyl', metavar='<bedmethyl.bed>',
        help='bedMethyl file containing methylation data')
    parser_promoter.add_argument('--upstream-flank', metavar='<int>', type=int,
        default=2000, help='length of upstream flank in bp (default: 2000)')
    parser_promoter.add_argument('--downstream-flank', metavar='<int>',
        type=int, default=0, help='length of upstream flank in bp (default: 0)')
    parser_promoter.add_argument('--chromosomes', metavar='<X>', nargs='+',
        help='chromosomes to include')
    parser_promoter.add_argument('--cytosines', action='store_true',
        help='output a column with number of cytosines in each promoter')
    parser_promoter.add_argument('--coverage', action='store_true',
        help='output a column of coverage for each promoter')
    parser_promoter.add_argument('--min-coverage', metavar='<int>',
        type=int, default=1, help='minimum coverage to include promoter (default: 1)')
    parser_promoter.add_argument('--bins', action='store_true',
        help='output a column binning promoters by discrete methylation level')
    parser_promoter.add_argument('--levels', metavar='<Level>',
        nargs='+', default=['Min', 'Low', 'Mid', 'High', 'Max'],
        help='discrete methylation levels')
    parser_promoter.add_argument('--hist', metavar='<file.{pdf,png,svg}>',
        help='generate histogram of methylation levels')
    parser_promoter.add_argument('--hist-log', action='store_true',
        help='use a log scale for  histogram')
    parser_promoter.add_argument('--palette', metavar='<palette>',
        default='mako_r', help='name of seaborn color palette (default: mako_r)')

    parser_gene_body = subparsers.add_parser('gene-body',
        help='quantify gene body methylation')
    parser_gene_body.set_defaults(func=_gene_body_methylation)
    parser_gene_body.add_argument('features', metavar='<features.gff3>',
        help='gff3 file of genomic features')
    parser_gene_body.add_argument('bedmethyl', metavar='<bedmethyl.bed>',
        help='bedMethyl file containing methylation data')
    parser_gene_body.add_argument('--upstream-flank', metavar='<int>', type=int,
        default=0, help='length of upstream flank in bp (default: 0)')
    parser_gene_body.add_argument('--downstream-flank', metavar='<int>',
        type=int, default=0, help='length of upstream flank in bp (default: 0)')
    parser_gene_body.add_argument('--chromosomes', metavar='<X>', nargs='+',
        help='chromosomes to include')
    parser_gene_body.add_argument('--cytosines', action='store_true',
        help='output a column with number of cytosines in each gene')
    parser_gene_body.add_argument('--coverage', action='store_true',
        help='output a column of coverage for each gene')
    parser_gene_body.add_argument('--min-coverage', metavar='<int>',
        type=int, default=1, help='minimum coverage to include gene (default: 1)')
    parser_gene_body.add_argument('--bins', action='store_true',
        help='output  a column binning genes by discrete methylation level')
    parser_gene_body.add_argument('--levels', metavar='<Level>',
        nargs='+', default=['Min', 'Low', 'Mid', 'High', 'Max'],
        help='discrete methylation levels')
    parser_gene_body.add_argument('--hist', metavar='<file.{pdf,png,svg}>',
        help='generate histogram of methylation levels')
    parser_gene_body.add_argument('--hist-log', action='store_true',
        help='use a log scale for  histogram')
    parser_gene_body.add_argument('--palette', metavar='<palette>',
        default='mako_r', help='name of seaborn color palette (default: mako_r)')
    
    parser_plot = subparsers.add_parser('plot',
        help='plot methylation across chromosomes')
    parser_plot.set_defaults(func=_plot)
    parser_plot.add_argument('bedmethyl', metavar='<bedmethyl.bed>',
        nargs='+', help='bedMethyl file containing methylation data')
    parser_plot.add_argument('output', metavar='<output.{pdf,png,svg}>',
        help='path to output file')
    parser_plot.add_argument('--reference', metavar='<reference.fa>',
        help='reference genome')
    parser_plot.add_argument('--chromosomes', metavar='<X>', nargs='+',
        help='chromosomes to plot')
    parser_plot.add_argument('--groups', metavar='<"Group">', nargs='+',
        help='list of groups for provided bedmethyl files (default: 0)')
    parser_plot.add_argument('--title', metavar='<"Plot title">',
        default='Methylation', help='set the title for the plot')
    parser_plot.add_argument('--x-label', metavar='<"Label">',
        default='Chromosome', help='set x-axis label for the plot (default: Chromosome)')
    parser_plot.add_argument('--legend', action='store_true',
        help='include a legend with the plot')
    parser_plot.add_argument('--legend-title', metavar='<"Title">',
        default='Group', help='title of legend')
    parser_plot.add_argument('--bin-size', metavar='<int>', type=int, default=0,
        choices=(-2,-1,0,1,2),
        help=('Set bin size. The input <int> is converted to the bin size by '
              'the formula: 10^(<int>+6) bp. The default value is 0, i.e. '
              '1-megabase bins. (default: 0)'))
    parser_plot.add_argument('--width', metavar='<float>', type=float,
        default=8.0, help='set width of figure in inches (default: 8.0)')
    parser_plot.add_argument('--color-palette', metavar='<#color>', nargs='+',
        default=COLOR_PALETTE, help='color palette to use')
    parser_plot.add_argument('--alpha', metavar='<float>', type=float,
        default=0.5, help='transparency value for lines (default: 0.5)')
    
    parser_plot_genes = subparsers.add_parser('plot-genes',
        help='plot methylation profiles over genomic features')
    parser_plot_genes.set_defaults(func=_plot_genes)
    parser_plot_genes.add_argument('features', metavar='<features.gff3>',
        help='gff3 file of genomic features')
    parser_plot_genes.add_argument('bedmethyl', metavar='<bedmethyl.bed>',
        nargs='+', help='bedMethyl file containing methylation data')
    parser_plot_genes.add_argument('output', metavar='<output.{pdf,png,svg}>',
        help='path to output file')
    parser_plot_genes.add_argument('--groups', metavar='<"Group">', nargs='+',
        help='list of groups for provided bedmethyl files [0)')
    parser_plot_genes.add_argument('--flank', metavar='<int>', type=int,
        default=2000, help='size of flanking regions in bp [1000)')
    parser_plot_genes.add_argument('--smooth', action='store_true',
        help='draw a smoother plot')
    parser_plot_genes.add_argument('--confidence-interval', metavar='<int>',
        type=int, help='draw a confidence interval')
    parser_plot_genes.add_argument('--title', metavar='<"Plot title">',
        default='Methylation',
        help='set the title for the plot')
    parser_plot_genes.add_argument('--gene-bins', metavar='<gene_bins.json>',
        help='gene bins')
    parser_plot_genes.add_argument('--gene-levels', metavar='<Level>',
        nargs='+', default=['Min', 'Low', 'Mid', 'High', 'Max'],
        help='gene expression levels')
    parser_plot_genes.add_argument('--legend', action='store_true',
        help='include a legend with the plot')
    parser_plot_genes.add_argument('--legend-title', metavar='<Title>',
        default='Group', help='title of legend')
    parser_plot_genes.add_argument('--width', metavar='<float>', type=float,
        default=4.0, help='set width of figure in inches (default: 4.0)')
    parser_plot_genes.add_argument('--palette', metavar='<palette>',
        default='deep', help='name of seaborn color palette (default: deep)')
    parser_plot_genes.add_argument('--alpha', metavar='<float>', type=float,
        default=1.0, help='transparency value for lines (default: 1.0)')
    
    parser_plot_repeats = subparsers.add_parser('plot-repeats',
        help='plot methylation profiles over genomic features')
    parser_plot_repeats.set_defaults(func=_plot_repeats)
    parser_plot_repeats.add_argument('features', metavar='<features.gff3>',
        help='gff3 file of genomic features')
    parser_plot_repeats.add_argument('bedmethyl', metavar='<bedmethyl.bed>',
        nargs='+', help='bedMethyl file containing methylation data')
    parser_plot_repeats.add_argument('output', metavar='<output.{pdf,png,svg}>',
        help='path to output file')
    parser_plot_repeats.add_argument('--type', metavar='<"feature_type">',
        help='generate plot for a specific type of repeat')
    parser_plot_repeats.add_argument('--groups', metavar='<"Group">', nargs='+',
        help='list of groups for provided bedmethyl files (default: 0)')
    parser_plot_repeats.add_argument('--flank', metavar='<int>', type=int,
        default=500, help='size of flanking regions in bp (default: 500)')
    parser_plot_repeats.add_argument('--smooth', action='store_true',
        help='draw a smoother plot')
    parser_plot_repeats.add_argument('--confidence-interval', metavar='<int>',
        type=int, help='draw a confidence interval')
    parser_plot_repeats.add_argument('--title', metavar='<"Plot title">',
        default='Methylation',
        help='set the title for the plot')
    parser_plot_repeats.add_argument('--legend', action='store_true',
        help='include a legend with the plot')
    parser_plot_repeats.add_argument('--legend-title', metavar='<Title>',
        default='Group', help='title of legend (default: Group)')
    parser_plot_repeats.add_argument('--width', metavar='<float>', type=float,
        default=4.0, help='set width of figure in inches (default: 4.0)')
    parser_plot_repeats.add_argument('--palette', metavar='<palette>',
        default='deep', help='name of seaborn color palette (default: deep)')
    parser_plot_repeats.add_argument('--alpha', metavar='<float>', type=float,
        default=1, help='transparency value for lines (default: 1.0)')
    
    parser_intersect = subparsers.add_parser('intersect',
        help='intersect two or more bedmethyl files')
    parser_intersect.set_defaults(func=_intersect)
    parser_intersect.add_argument('bedmethyl', metavar='<bedmethyl.bed>',
        nargs='+', help='bedMethyl file containing methylation data')
    parser_intersect.add_argument('output_prefix', metavar='<output-prefix>',
        help='prefix for output files')
    parser_intersect.add_argument('--chromosomes', metavar='<X>',
        nargs='+', help='chromosomes to include')

    parser_clean = subparsers.add_parser('clean',
        help='clean up downloads for uninstall')
    parser_clean.set_defaults(func=clean)
    parser_clean.add_argument('--dirs', metavar='<directory/>',
        nargs='+', default=[PBCPG_DIR, DORADO_DIR, DORADO_MODEL_DIR],
        help=f"directories to clean up (default: {', '.join(set([PBCPG_DIR, DORADO_DIR, DORADO_MODEL_DIR]))})")

    return parser.parse_args()


def main():
    args = parse_arguments()
    args.func(args)



#===============================================================================
# plot.py
#===============================================================================

"""Plot methylation data across chromosomes"""




# Imports ======================================================================

from itertools import chain, groupby, accumulate, islice, cycle
from operator import itemgetter
from pybedtools import BedTool
from pyfaidx import Fasta

import pandas as pd
import seaborn as sns




# Constants ====================================================================

COLOR_PALETTE = sns.color_palette().as_hex()




# Functions ====================================================================

def get_chromosome_sizes_from_ref(reference: str):
    """Extract chromosome sizes from reference FASTA

    Parameters
    ----------
    reference : str
        path to reference FASTA file

    Returns
    -------
    DataFrame
        name and size of each chromosome
    """

    return pd.DataFrame(((k, len(v)) for k, v in Fasta(reference).items()),
                        columns=('name', 'size'))


def approx_chromosome_sizes_from_data(bedtools):
    """Approximate chromosome sizes from BED data

    Parameters
    ----------
    bedtools
        iterble of bedtools generated from methylbed files

    Returns
    -------
    DataFrame
        name and approximate size of each chromosome
    """

    return pd.DataFrame(((key, max(tuple(zip(*val))[1]))
        for key, val in groupby(sorted((k, (max(int(x[2]) for x in v)))
            for k, v in groupby((i.fields for bt in bedtools for i in bt),
                key=itemgetter(0))),
            key=itemgetter(0))), columns=('name', 'size'))


def generate_plotting_data(bedtools, groups, size, scale: float = 1,
                           shift: float = 0, bin_size: int = 0):
    """Construct rows of preprocessed data for the plotting data frame. Data
    are binned by rounding to the nearest bin coordinate, while bin coordinates
    are determined by the bin size parameter.

    Parameters
    ----------
    bedtools
        iterable of bedtools generated from methylbed data
    groups
        iterable of group names
    size
        chromosome size in bp
    scale
        ratio of chromosome size to mean chromosome size
    shift
        x-axis shift of this chromosome, for plots showing multiple chromosomes
        consecutively
    bin_size
        set bin size. The input <int> is converted to the bin size by the
        formula: 10^(<int>+6) bp. The default value is 0, i.e. 1-megabase bins.

    Yields
    ------
    tuple
        bin coordinate, value, and group ID of a methylation data point
    """

    for bt, g in zip(bedtools, groups):
        for i in bt:
            n_fields = len(i.fields)
            if n_fields == 10:
                chrom, _, pos, _, _, _, _, _, _, meth_col, group = i.fields + [g]
                meth = meth_col.split()[1]
            elif n_fields == 9:
                chrom, _, pos, meth, _, _, _, _, _, group = i.fields + [g]
            else:
                raise RuntimeError('Invalid methylation BED file')
            yield (min(round(int(pos), -6-bin_size), size)/size*scale + shift,
                float(meth), group, f'{group}_{chrom}')


def plot(bedmethyl, output, reference=None, chromosomes=['1'], groups=None,
         title: str = 'Methylation', legend: bool = False,
         legend_title: str = 'Group', bin_size: int = 0,
         width: float = 8.0, color_palette=COLOR_PALETTE, alpha: float = 0.5,
         x_label: str = 'Chromosome',):
    """Generate a plot of average methylation levels across one or more
    chromosomes

    Parameters
    ----------
    bedmethyl
        iterable of paths path to input bedmethyl files
    output
        path to output file (pdf, png, or svg)
    reference
        path to FASTA file for reference genome
    chromosomes
        iterable of chromosomes to include in plot
    groups
        iterable of group ids for input files
    title : str
        title for plot
    legend : bool
        if true, draw a levend for the plot
    legend_title : str
        title for plot legend
    bin_size : int
        set bin size. The input <int> is converted to the bin size by the
        formula: 10^(<int>+6) bp. The default value is 0, i.e. 1-megabase bins.
    width : float
        width of plot in inches
    color_palette
        color palette for lines
    alpha
        transparency of lines
    """

    if not groups:
        groups = list(range(len(bedmethyl)))
    methyl = tuple(BedTool(bed) for bed in bedmethyl)
    sizes = (get_chromosome_sizes_from_ref(reference) if reference
             else approx_chromosome_sizes_from_data(methyl))
    sizes.index = sizes.name
    sizes = sizes.loc[chromosomes, 'size']
    scales = sizes / sizes.mean()
    shifts = pd.Series(accumulate(chain((0,),scales[:-1])), index=scales.index)

    plotting_data = pd.DataFrame(
        chain.from_iterable((generate_plotting_data(
                (m.intersect(BedTool(((chrom, 0, size),))) for m in methyl),
                groups, size, scale=scale, shift=shift, bin_size=bin_size)
            for chrom, size, scale, shift
            in zip(chromosomes, sizes, scales, shifts))),
        columns=(x_label, 'Methylation level (%)', legend_title,
                 f'{legend_title}_chrom'))

    palette = tuple(islice(cycle(color_palette[:len(groups)]),
                                 len(chromosomes) * len(groups)))
    ax = sns.lineplot(x=x_label, y='Methylation level (%)',
                      hue=f'{legend_title}_chrom', data=plotting_data,
                      errorbar=None, linewidth=3, palette=palette,
                      alpha=alpha, legend='auto' if legend else False)
    ax.set_title(title)
    if (len(chromosomes) == 1) and (sizes is not None):
        xticks = ax.get_xticks()[1:-1]
        xlabels=tuple(f"{x*sizes.loc[chromosomes[0]]/1e6:.1f}" for x in xticks)
    else:
        xticks = shifts
        xlabels = chromosomes
    ax.set_xticks(xticks)
    ax.set_xticklabels(xlabels, rotation=30, ha='right')
    if legend:
        leg = ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left',
                        borderaxespad=0, title=legend_title)
        for line in leg.get_lines():
            line.set_linewidth(3)
            line.set_alpha(alpha)
    fig = ax.get_figure()
    fig.set_figheight(3)
    fig.set_figwidth(width)
    fig.tight_layout()
    fig.savefig(output)
    fig.clf()



#===============================================================================
# plot_genes.py
#===============================================================================

from itertools import chain
from pybedtools import BedTool

import pandas as pd
import seaborn as sns
import json
import gff2bed

from statistics import median
from loreme.parse_gff import generate_flank, generate_feature_body
from loreme.plot_bedtools import generate_plotting_data



def plot_genes(features, bedmethyl, output, groups=None, flank: int = 2000,
               smooth: bool = False, title: str = 'Methylation', gene_bins=None,
               gene_levels=['Min', 'Low', 'Mid', 'High', 'Max'],
               confidence_interval=None,
               legend=False, legend_title='Group', width: float = 4.0,
               palette='deep', alpha=1):
    """Plot methylation profile of genes across the genome

    Parameters
    ----------
    features
        path to GFF3 file containing gene coordinates
    bedmethyl
        iterable of paths to bedmethyl files
    output
        path to output file (pdf, png, or svg)
    groups
        iterable of group names for input files
    flank : int
        size of flanking regions to plot, in bp
    smooth : bool
        if True, draw a smoother plot
    title : str
        title for plot
    gene_bins
        path to json or tsv file providing genes binned by expression
    gene_levels
        iterable of names for discrete gene expression levels
    confidence_interval
        int from 1-99 giving % size of confidence interval, or None for
        no confidence interval
    legend : bool
        if True, draw a legend
    legend title : str
        title for plot legend
    width : float
        width of plot in inches
    palette : str
        color palette for lines
    alpha : float
        transparency of lines
    """

    if not groups:
        groups = list(range(len(bedmethyl)))
    genes = pd.DataFrame(gff2bed.parse(features),
        columns=('seqid', 'start', 'end', 'strand', 'attributes'))
    genes.index = (attr['ID'] for attr in genes['attributes'])
    gene_size=median(genes['end'] - genes['start'])
    xticks = (0, 3*flank/(gene_size+2*flank),
              3*(gene_size+flank)/(gene_size+2*flank), 3)
    upstream = BedTool(tuple(f for f in generate_flank(genes, side='up', flank=flank) if f[1]>0))
    downstream = BedTool(tuple(f for f in generate_flank(genes, side='down', flank=flank) if f[1]>0))
    gene_body = BedTool(tuple(f for f in generate_feature_body(genes) if f[1]>0))
    methyl = tuple(BedTool(bed) for bed in bedmethyl)

    methyl_upstream = (m.intersect(upstream, wo=True) for m in methyl)
    methyl_gene_body = (m.intersect(gene_body, wo=True) for m in methyl)
    methyl_downstream = (m.intersect(downstream, wo=True) for m in methyl)

    plotting_data = pd.DataFrame(chain(
        generate_plotting_data(methyl_upstream, groups, scale=xticks[1],
            smooth=smooth),
        generate_plotting_data(methyl_gene_body, groups,
            scale=xticks[2] - xticks[1], shift=xticks[1], smooth=smooth),
        generate_plotting_data(methyl_downstream, groups,
            scale=xticks[3] - xticks[2], shift=xticks[2], smooth=smooth)),
        columns=('Relative position', 'Methylation level (%)', legend_title,
                 'ID'))
    
    if gene_bins:
        print('applying bins')
        if gene_bins.endswith('.json'):
            with open(gene_bins, 'r') as f:
                bins = json.load(f)
        else:
            bins = pd.read_table(gene_bins, header=None, index_col=0,
                                 squeeze=True)
        plotting_data[legend_title] = tuple(bins.get(f'{gene_id}')
            for gene_id in plotting_data['ID'])
        plotting_data.dropna()

    if gene_bins:
        palette = sns.color_palette(palette,
            len(plotting_data[legend_title].unique()))[-1*len(gene_levels):]
    else:
        palette = sns.color_palette(palette, n_colors=len(groups))
    ax = sns.lineplot(x='Relative position', y='Methylation level (%)',
                      hue=legend_title, data=plotting_data,
                      errorbar=confidence_interval, linewidth=3,
                      hue_order=(gene_levels if gene_bins else None),
                      palette=palette, legend='auto' if legend else False,
                      alpha=alpha)
    ax.set_title(title)
    ax.set_xticks(xticks)
    ax.set_xticklabels((f'-{int(flank/1e3)}kb', 'TSS', 'TTS',
                        f'+{int(flank/1e3)}kb'),
                       rotation=30, ha='right')
    ax.axvline(x=xticks[1], color='lightgray', linestyle='dashed')
    ax.axvline(x=xticks[2], color='lightgray', linestyle='dashed')
    if legend:
        leg = ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left',
                        borderaxespad=0, title=legend_title)
        for line in leg.get_lines():
            line.set_linewidth(3)
    fig = ax.get_figure()
    fig.set_figheight(3)
    fig.set_figwidth(width)
    fig.tight_layout()
    fig.savefig(output)
    fig.clf()



import os
import os.path
import json

PBCPG_DIR = os.environ.get('LOREME_PBCPG_DIR',
    os.path.join(os.path.dirname(__file__)))
PBCPG_PATH = os.path.join(PBCPG_DIR, 'pb-CpG-tools-v2.3.1-x86_64-unknown-linux-gnu',
                         'bin', 'aligned_bam_to_cpg_scores')
PBCPG_MODEL = os.path.join(PBCPG_DIR, 'pb-CpG-tools-v2.3.1-x86_64-unknown-linux-gnu',
                               'models', 'pileup_calling_model.v1.tflite')
PBCPG_URL = 'https://github.com/PacificBiosciences/pb-CpG-tools/releases/download/v2.3.1/pb-CpG-tools-v2.3.1-x86_64-unknown-linux-gnu.tar.gz'
PB_EXAMPLE_DATA_URLS = (
    'https://downloads.pacbcloud.com/public/dataset/HG002-CpG-methylation-202202/m64011_190830_220126.hifi_reads.bam',
    'https://downloads.pacbcloud.com/public/dataset/HG002-CpG-methylation-202202/m64011_190901_095311.hifi_reads.bam',
    'https://downloads.pacbcloud.com/public/dataset/HG002-CpG-methylation-202202/m64012_190920_173625.hifi_reads.bam',
    'https://downloads.pacbcloud.com/public/dataset/HG002-CpG-methylation-202202/m64012_190921_234837.hifi_reads.bam',
)
HG38_FTP = 'ftp.ncbi.nlm.nih.gov'
HG38_GENOME_PATH = 'genomes/all/GCA/000/001/405/GCA_000001405.15_GRCh38/seqs_for_alignment_pipelines.ucsc_ids/GCA_000001405.15_GRCh38_no_alt_analysis_set.fna.gz'
HG38_ANNOT_PATH = 'genomes/all/GCA/000/001/405/GCA_000001405.15_GRCh38/seqs_for_alignment_pipelines.ucsc_ids/GCA_000001405.15_GRCh38_full_analysis_set.refseq_annotation.gff.gz'
ONT_EXAMPLE_DATA_URL = 'https://salk-tm-pub.s3.us-west-2.amazonaws.com/LoReMe_example/Kazu_090722_95etoh_22C_6w_SRE1_deep.pod5'
KAZU_090722_GENOME_URL = 'https://salk-tm-pub.s3.us-west-2.amazonaws.com/LoReMe_example/Kazu_090722.softmasked.fasta.gz'
KAZU_090722_ANNOT_URL = 'https://salk-tm-pub.s3.us-west-2.amazonaws.com/LoReMe_example/Kazu_090722.primary_high_confidence.gff3.gz'
EXAMPLE_DATA_DIR = os.environ.get('LOREME_EXAMPLE_DATA_DIR', os.path.dirname(__file__))

DORADO_DIR = os.environ.get('LOREME_DORADO_DIR',
    os.path.join(os.path.dirname(__file__)))
DORADO_CONFIG = os.environ.get('LOREME_DORADO_CONFIG',
    os.path.join(os.path.dirname(__file__), 'dorado-config.json'))
if os.path.isfile(DORADO_CONFIG):
    with open(DORADO_CONFIG, 'r') as f:
        dorado_config = json.load(f)
else:
    dorado_config = {}
DORADO_PLATFORM = os.environ.get('LOREME_DORADO_PLATFORM', dorado_config.get('platform'))
DORADO_PATH = {
    'linux-x64': os.path.join(DORADO_DIR, 'dorado-0.3.2-linux-x64', 'bin', 'dorado'),
    'linux-arm64': os.path.join(DORADO_DIR, 'dorado-0.3.2-linux-arm64', 'bin', 'dorado'),
    'osx-arm64': os.path.join(DORADO_DIR, 'dorado-0.3.2-osx-arm64', 'bin', 'dorado'),
    'win64': os.path.join(DORADO_DIR, 'dorado-0.3.2-win64', 'bin', 'dorado')
}
DORADO_PATH_024 = {
    'linux-x64': os.path.join(DORADO_DIR, 'dorado-0.2.4-linux-x64', 'bin', 'dorado'),
    'linux-arm64': os.path.join(DORADO_DIR, 'dorado-0.2.4-linux-arm64', 'bin', 'dorado'),
    'osx-arm64': os.path.join(DORADO_DIR, 'dorado-0.2.4-osx-arm64', 'bin', 'dorado'),
    'win64': os.path.join(DORADO_DIR, 'dorado-0.2.4-win64', 'bin', 'dorado')
}
DORADO_URL = {
    'linux-x64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.3.2-linux-x64.tar.gz',
    'linux-arm64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.3.2-linux-arm64.tar.gz',
    'osx-arm64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.3.2-osx-arm64.tar.gz',
    'win64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.3.2-win64.zip'
}
DORADO_URL_024 = {
    'linux-x64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.2.4-linux-x64.tar.gz',
    'linux-arm64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.2.4-linux-arm64.tar.gz',
    'osx-arm64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.2.4-osx-arm64.tar.gz',
    'win64': 'https://cdn.oxfordnanoportal.com/software/analysis/dorado-0.2.4-win64.zip'
}
DORADO_MODEL_DIR = os.environ.get('LOREME_DORADO_MODEL_DIR',
    os.path.join(os.path.dirname(__file__)))
MODKIT_URL = 'https://github.com/nanoporetech/modkit/releases/download/v0.1.8/modkit_v0.1.8_centos7_x86_64.tar.gz'
MODKIT_DIR = os.environ.get('LOREME_MODKIT_DIR',
    os.path.join(os.path.dirname(__file__)))
MODKIT_PATH = os.path.join(MODKIT_DIR, 'dist', 'modkit')



"""LoReMe (Long Read Methylaton) is a Python package facilitating analysis of
DNA methylation signals from `Pacific Biosciences <https://www.pacb.com/technology/hifi-sequencing/>`_
or `Oxford Nanopore <https://nanoporetech.com/applications/dna-nanopore-sequencing>`_
long read sequencing data.

It consists of an API and CLI for three distinct applications:

1. Pacific Biosciences data processing. PB reads in SAM/BAM format are aligned
to a reference genome with the special-purpose aligner `pbmm2 <https://github.com/PacificBiosciences/pbmm2>`_ ,
a modified version of `minimap2 <https://lh3.github.io/minimap2/>`_ .
Methylation calls are then piled up from the aligned reads with `pb-CpG-tools <https://github.com/PacificBiosciences/pb-CpG-tools>`_ .

2. Oxford nanopore basecalling. ONT reads are optionally converted from FAST5
to `POD5 <https://github.com/nanoporetech/pod5-file-format>`_ format, then
basecalled and aligned to a reference with `dorado <https://github.com/nanoporetech/dorado>`_ 
(dorado alignment also uses minimap2 under the hood), and finally piled up with
`modkit <https://github.com/nanoporetech/modkit/>`_ .

3. Postprocessing and QC of methylation calls. Several functions are available
to generate diagnostic statistics and plots.

Other tools of interest: `methylartist <https://github.com/adamewing/methylartist>`_ , `modbamtools <https://github.com/rrazaghi/modbamtools>`_  (`modbamtools docs <https://rrazaghi.github.io/modbamtools/>`_), `methplotlib <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7214038/>`_
"""

from loreme.version import __version__
from loreme.env import PBCPG_MODEL
from loreme.download import download_pbcpg, download_dorado
from loreme.check_gpu_availability import check_gpu_availability
from loreme.check_tags import check_tags
from loreme.pbcpg import pbcpg_align_bam, pbcpg_align_bams, pbcpg_pileup
from loreme.dorado import dorado_basecall



import os
import subprocess
import pysam
from loreme.env import PBCPG_PATH, PBCPG_MODEL

def pbcpg_align_bam(ref: str, in_bam: str, out_bam: str, threads: int = 1,
              memory_mb: int = 768):
    """Run pbmm2 to align a BAM file to a FASTA reference

    Parameters
    ----------
    ref : str
        path to reference FASTA
    in_bam : str
        path to input BAM
    out_bam : str
        path to output BAM
    threads : int
        number of threads to use
    memory_mb : int
        megabytes of memory to use
    """
    subprocess.run(('pbmm2', 'align', '--preset', 'HIFI', '--sort',
                    '-j', f'{threads}', '--sort-memory', f'{memory_mb}M',
                    ref, in_bam, out_bam))


def pbcpg_align_bams(ref: str, in_bams, out_bam: str, threads: int = 1,
              memory_mb: int = 768):
    """Run pbmm2 to align a BAM file to a FASTA reference

    Parameters
    ----------
    ref : str
        path to reference FASTA
    in_bams : str
        paths to input BAMs
    out_bam : str
        path to output BAM
    threads : int
        number of threads to use
    memory_mb : int
        megabytes of memory to use
    """
    
    if len(in_bams) == 1:
        pbcpg_align_bam(ref, in_bams[0], out_bam, threads=threads,
                  memory_mb=memory_mb)
    else:
        for n, in_bam in enumerate(in_bams):
            pbcpg_align_bam(ref, in_bam, f'{out_bam[:-4]}_{n}.bam', threads=threads,
                    memory_mb=memory_mb)
        pysam.merge('-o', out_bam,
                    *(f'{out_bam[:-4]}_{n}.bam' for n in range(len(in_bams))))
        pysam.index(out_bam)
        for n in range(len(in_bams)):
            os.remove(f'{out_bam[:-4]}_{n}.bam')
            os.remove(f'{out_bam[:-4]}_{n}.bam.bai')

def pbcpg_pileup(bam, output_prefix, hap_tag: str = 'HP',
                  min_coverage: int = 4, min_mapq: int = 1,
                  model: str = PBCPG_MODEL, threads: int = 1):
    subprocess.run((PBCPG_PATH, '--bam', bam,
        '--output-prefix', output_prefix,
        '--hap-tag', hap_tag,
        '--min-coverage', str(min_coverage),
        '--min-mapq', str(min_mapq),
        '--model', model,
        '--threads', str(threads)))



#===============================================================================
# methylation_hist.py
#===============================================================================

import seaborn as sns

def methylation_hist(df, output_file, palette='mako_r', log_scale=False):
    """Draw a histogram of methylation levels

    Parameters
    ----------
    df
        input data frame of methylation levels
    output_file
        path to output file
    palette
        color palette for discrete methylation levels
    xlog : bool
        if True, draw a log scale on the x-axis
    """

    ax = sns.histplot(data=df, x='Methylation level (%)',
                    hue='Discrete level', fill=True, palette=palette,
                    element='step', multiple='stack', log_scale=log_scale)
    fig = ax.get_figure()
    fig.set_figwidth(7)
    fig.set_figheight(3)
    fig.tight_layout()
    fig.savefig(output_file)
    fig.clf()



#===============================================================================
# plot_repeats.py
#===============================================================================

from itertools import chain
from pybedtools import BedTool

import pandas as pd
import seaborn as sns
import gff2bed

from statistics import median
from loreme.parse_gff import (generate_flank,
                                          generate_feature_body)
from loreme.plot_bedtools import generate_plotting_data


def plot_repeats(features, bedmethyl, output, type=None, groups=None,
               flank: int = 500, smooth: bool = False,
               title: str = 'Methylation', confidence_interval=None,
               legend=False, legend_title='Group', width: float = 4.0,
               palette='deep', alpha=1):
    """Plot methylation profile of repeats across the genome

    Parameters
    ----------
    features
        path to GFF3 file containing repeat coordinates
    bedmethyl
        iterable of paths to bedmethyl files
    output
        path to output file (pdf, png, or svg)
    type : str
        type of repeat to plot, or None to plot all types in a single plot
    groups
        iterable of group names for input files
    flank : int
        size of flanking regions to plot, in bp
    smooth : bool
        if True, draw a smoother plot
    title : str
        title for plot
    confidence_interval
        int from 1-99 giving % size of confidence interval, or None for
        no confidence interval
    legend : bool
        if True, draw a legend
    legend title : str
        title for plot legend
    width : float
        width of plot in inches
    palette : str
        color palette for lines
    alpha : float
        transparency of lines
    """
    
    if not groups:
        groups = list(range(len(bedmethyl)))
    repeats = pd.DataFrame(gff2bed.parse(features, type=type),
        columns=('seqid', 'start', 'end', 'strand', 'attributes'))
    repeats.index = (attr['ID'] for attr in repeats['attributes'])
    repeat_size=median(repeats['end'] - repeats['start'])
    xticks = (0, 3*flank/(repeat_size+2*flank),
              3*(repeat_size+flank)/(repeat_size+2*flank), 3)
    upstream = BedTool(tuple(f for f in generate_flank(repeats, side='up', flank=flank) if f[1]>0))
    downstream = BedTool(tuple(f for f in generate_flank(repeats, side='down', flank=flank) if f[1]>0))
    repeat_body = BedTool(tuple(f for f in generate_feature_body(repeats) if f[1]>0))
    methyl = tuple(BedTool(bed) for bed in bedmethyl)

    methyl_upstream = (m.intersect(upstream, wo=True) for m in methyl)
    methyl_repeat_body = (m.intersect(repeat_body, wo=True) for m in methyl)
    methyl_downstream = (m.intersect(downstream, wo=True) for m in methyl)

    plotting_data = pd.DataFrame(chain(
        generate_plotting_data(methyl_upstream, groups, scale=xticks[1],
            smooth=smooth),
        generate_plotting_data(methyl_repeat_body, groups,
            scale=xticks[2] - xticks[1], shift=xticks[1], smooth=smooth),
        generate_plotting_data(methyl_downstream, groups,
            scale=xticks[3] - xticks[2], shift=xticks[2], smooth=smooth)),
        columns=('Relative position', 'Methylation level (%)', legend_title,
                 'ID'))

    palette = sns.color_palette(palette, n_colors=len(groups))
    ax = sns.lineplot(x='Relative position', y='Methylation level (%)',
                      hue=legend_title, data=plotting_data,
                      ci=confidence_interval, linewidth=3, palette=palette,
                      legend='auto' if legend else False, alpha=alpha)
    ax.set_title(title)
    ax.set_xticks(xticks)
    ax.set_xticklabels((f'-{int(flank)}bp', 'First', 'Last',
                        f'+{int(flank)}bp'),
                       rotation=30, ha='right')
    ax.axvline(x=xticks[1], color='lightgray', linestyle='dashed')
    ax.axvline(x=xticks[2], color='lightgray', linestyle='dashed')
    if legend:
        leg = ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left',
                        borderaxespad=0, title=legend_title)
        for line in leg.get_lines():
            line.set_linewidth(3)
    fig = ax.get_figure()
    fig.set_figheight(3)
    fig.set_figwidth(width)
    fig.tight_layout()
    fig.savefig(output)
    fig.clf()
    



#===============================================================================
# plot_bedtools.py
#===============================================================================

def generate_feature_data(bedtool, chromosomes=None):
    """Generate (as tuples) entries from a bedtool object containing
    methylation data

    Parameters
    ----------
    bedtool
        a BedTool object
    chromosomes
        iterable of chromosomes to include in results, or None to include
        all chromosomes

    Yield
    -----
    tuple
        
    """

    if chromosomes is not None:
        chromosomes = set(chromosomes)
    for i in bedtool:
        n_fields = len(i.fields)
        if n_fields == 16:
            chrom, _, _, _, coverage, _, _, _, _, meth_col, _, start, end, strand, gene, _ = i.fields
            meth = meth_col.split()[1]
        elif n_fields == 15:
            chrom, _, _, meth, _, coverage, _, _, _, _, start, end, strand, gene, _ = i.fields
        else:
            raise RuntimeError('Invalid methylation BED file')
        if (chromosomes is None) or (chrom in chromosomes):
            yield chrom, int(start), int(end), gene, strand, 1, int(coverage), float(meth)


def generate_plotting_data(bedtools, groups, scale: float = 1,
                           shift: float = 0, smooth: bool = False):
    """Generate methylation data converted to plotting coordinates for gene
    or TE profile plots

    Parameters
    ----------
    bedtools
        iterable of bedtools containing methylation data
    groups
        iterable of group names for input datasets
    scale : float
        ratio of chromosome size to mean chromosome size
    shift : float
        x-axis shift of this chromosome, for plots showing multiple chromosomes
        consecutively
    smooth : bool
        if True, draw a smoother plot

    Yield
    -----
    tuple
        
    """

    for bt, g in zip(bedtools, groups):
        for i in bt:
            n_fields = len(i.fields)
            if n_fields == 16:
                _, _, pos, _, _, _, _, _, _, meth_col, _, start, end, strand, index, group = i.fields
                meth = meth_col.split()[1]
            elif n_fields == 15:
                _, _, pos, meth, _, _, _, _, _, _, start, end, strand, index, group = i.fields+[g]
            else:
                raise RuntimeError('Invalid methylation BED file')
            yield (round((int(pos)-int(start) if strand == '+' else int(end)-int(pos))
         / (int(end)-int(start)), 2-smooth)*scale + shift, float(meth), group, index)



import pysam
from itertools import islice

def check_tags(bam, n_reads: int = 1000):
    samfile = pysam.AlignmentFile(bam, 'r', check_sq=False)
    has_tags = any(((read.has_tag('MM') or read.has_tag('Mm'))
                    and (read.has_tag('ML') or read.has_tag('Ml'))
                    for read in islice(samfile.fetch(until_eof=True), n_reads)))
    samfile.close()
    return has_tags



#===============================================================================
# merge.py
#===============================================================================

"""merge two bedmethyl files"""

# Imports ======================================================================

from pybedtools import BedTool
from itertools import chain
import numpy as np




# Functions ====================================================================


def merge_bedmethyl(bedmethyl, chromosomes=None):
    """Merge two or more bedmethyl files

    Parameters
    ----------
    bedmethyl
        iterable of paths to bedmethyl formatted files of methylation results
    chromosomes
        iterable of chromosomes to include in output
    """
    
    bedtools = tuple(BedTool(bed) for bed in bedmethyl)
    chromosomes=set(chromosomes) if chromosomes else None
    merged = BedTool((i for i in chain(*bedtools)
        if (chromosomes is None) or (i.chrom in chromosomes))).sort().merge(
        s=True,
        c='4,5,6,7,8,9,10,11',
        o='first,sum,first,first,last,first,collapse,collapse')
    for i in merged:
        chrom, start, stop = i.fields[:3]
        score = min(int(i.fields[4]), 1000)
        strand, thick_start, thick_stop = i.fields[5:8]
        coverage, methyl = (tuple(x) for x in zip(
            *((int(c), float(m))
            for c, m in zip(i.fields[9].split(','), i.fields[10].split(','))
            if (m != '.'))))
        if sum(coverage) == 0:
            continue
        print(chrom, start, stop, '.', score, strand, thick_start,
                thick_stop, '0,0,0', sum(coverage),
                np.average(methyl, weights=coverage), sep='\t')



import urllib3
import shutil
import os
import os.path
import ftplib
import gzip
import tarfile
import subprocess
import json
from itertools import product
from loreme.env import (PBCPG_URL, PBCPG_DIR, PB_EXAMPLE_DATA_URLS, EXAMPLE_DATA_DIR, HG38_FTP,
                       HG38_GENOME_PATH, HG38_ANNOT_PATH, DORADO_DIR, DORADO_URL, DORADO_URL_024,
                       DORADO_MODEL_DIR, DORADO_CONFIG, DORADO_PATH, DORADO_PATH_024, MODKIT_URL,
                       MODKIT_DIR, ONT_EXAMPLE_DATA_URL, KAZU_090722_GENOME_URL,
                       KAZU_090722_ANNOT_URL)

def download_pbcpg(directory: str = PBCPG_DIR):
    """Download pb-CpG-tools

    Parameters
    ----------
    directory : str
        Destination directory for pb-CpG-tools
    """

    dest_tarfile = os.path.join(directory, 'pb-CpG-tools-v2.3.1-x86_64-unknown-linux-gnu.tar.gz')
    print(f'downloading to {dest_tarfile}')
    if os.path.exists(dest_tarfile):
        raise RuntimeError(f'a file already exists at {dest_tarfile}')
    http = urllib3.PoolManager()
    with http.request('GET', PBCPG_URL, preload_content=False) as r, open(dest_tarfile, 'wb') as f:
        shutil.copyfileobj(r, f)
    with tarfile.open(dest_tarfile) as tar:
        tar.extractall(path=directory)


def download_pb_example(directory: str = EXAMPLE_DATA_DIR, n_samples: int = 1):
    """Download an example datatset

    Parameters
    ----------
    directory : str
        Destination directory for example data
    n_samples : int
        Number of samples to download, max 4 [1]
    """

    if n_samples > 4:
        raise RuntimeError('n_samples parameter must be <= 4')
    hg38_genome_local_path = os.path.join(directory, os.path.basename(HG38_GENOME_PATH).replace('fna.gz', 'fa'))
    hg38_annot_local_path = os.path.join(directory, os.path.basename(HG38_ANNOT_PATH))
    if os.path.exists(hg38_genome_local_path):
        print(f'destination {hg38_genome_local_path} already exists')
    else:
        ftp = ftplib.FTP(HG38_FTP)
        ftp.login()
        with open(f'{hg38_genome_local_path}.gz', 'wb') as f:
            ftp.retrbinary(f'RETR {HG38_GENOME_PATH}', f.write)
    with gzip.open(f'{hg38_genome_local_path}.gz', 'rb') as f_in:
        with open(hg38_genome_local_path, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    os.remove(f'{hg38_genome_local_path}.gz')
    if os.path.exists(hg38_annot_local_path):
        print(f'destination {hg38_annot_local_path} already exists')
    else:
        ftp = ftplib.FTP(HG38_FTP)
        ftp.login()
        with open(hg38_annot_local_path, 'wb') as f:
            ftp.retrbinary(f'RETR {HG38_ANNOT_PATH}', f.write)
    http = urllib3.PoolManager()
    for example_data_url in PB_EXAMPLE_DATA_URLS[:n_samples]:
        file_path = os.path.join(directory, os.path.basename(example_data_url))
        if os.path.exists(file_path):
            print(f'destination {file_path} already exists')
            continue
        with http.request('GET', example_data_url, preload_content=False) as r, open(file_path, 'wb') as out_file:
            shutil.copyfileobj(r, out_file)


def download_ont_example(directory: str = EXAMPLE_DATA_DIR):
    """Download an example datatset

    Parameters
    ----------
    directory : str
        Destination directory for example data
    """

    kazu_genome_local_path = os.path.join(directory, os.path.basename(KAZU_090722_GENOME_URL).replace('fna.gz', 'fa'))
    kazu_annot_local_path = os.path.join(directory, os.path.basename(KAZU_090722_ANNOT_URL))
    ont_example_data_local_path = os.path.join(directory, os.path.basename(ONT_EXAMPLE_DATA_URL))
    http = urllib3.PoolManager()
    if os.path.exists(kazu_genome_local_path):
        print(f'destination {kazu_genome_local_path} already exists')
    else:
        with http.request('GET', KAZU_090722_GENOME_URL, preload_content=False) as r, \
            open(kazu_genome_local_path, 'wb') as out_file:
            shutil.copyfileobj(r, out_file)
    if os.path.exists(kazu_annot_local_path):
        print(f'destination {kazu_annot_local_path} already exists')
    else:
        with http.request('GET', KAZU_090722_ANNOT_URL, preload_content=False) as r, \
            open(kazu_annot_local_path, 'wb') as out_file:
            shutil.copyfileobj(r, out_file)
    if os.path.exists(ont_example_data_local_path):
        print(f'destination {ont_example_data_local_path} already exists')
    else:
        with http.request('GET', ONT_EXAMPLE_DATA_URL, preload_content=False) as r, \
            open(ont_example_data_local_path, 'wb') as out_file:
            shutil.copyfileobj(r, out_file)


def download_dorado(pfm, directory=DORADO_DIR, model_dir=DORADO_MODEL_DIR,
                    dorado_config=DORADO_CONFIG):
    """Download dorado

    Parameters
    ----------
    pfm : str
        platform string (linux-x86, linux-arm64, osx-arm64, win64)
    directory : str
        Destination directory for dorado
    model_dir : str
        destination directory for dorado models
    """

    if pfm not in {'linux-x64', 'linux-arm64', 'osx-arm64', 'win64'}:
        raise RuntimeError('invalid platform choice')
    dest_tarfile = os.path.join(directory, f'dorado-0.3.2-{pfm}.tar.gz')
    # dest_tarfile_024 = os.path.join(directory, f'dorado-0.2.4-{pfm}.tar.gz')
    print(f'downloading to {dest_tarfile}')
    if os.path.exists(dest_tarfile):
        raise RuntimeError(f'a file already exists at {dest_tarfile}')
    http = urllib3.PoolManager()
    with http.request('GET', DORADO_URL[pfm], preload_content=False) as r, open(dest_tarfile, 'wb') as f:
        shutil.copyfileobj(r, f)
    with tarfile.open(dest_tarfile) as tar:
        tar.extractall(path=directory)
    # with http.request('GET', DORADO_URL_024[pfm], preload_content=False) as r, open(dest_tarfile_024, 'wb') as f:
    #     shutil.copyfileobj(r, f)
    # with tarfile.open(dest_tarfile_024) as tar:
    #     tar.extractall(path=directory)
    print(f'downloading models to {model_dir}')
    for speed, accuracy, in product((260, 400), ('fast', 'hac', 'sup')):
        subprocess.run((DORADO_PATH[pfm], 'download', '--directory', model_dir,
            '--model', f'dna_r10.4.1_e8.2_{speed}bps_{accuracy}@v4.1.0'))
        subprocess.run((DORADO_PATH[pfm], 'download', '--directory', model_dir,
            '--model', f'dna_r10.4.1_e8.2_{speed}bps_{accuracy}@v4.1.0_5mCG_5hmCG@v2'))
    for accuracy in 'fast', 'hac', 'sup':
        subprocess.run((DORADO_PATH[pfm], 'download', '--directory', model_dir,
            '--model', f'dna_r10.4.1_e8.2_400bps_{accuracy}@v4.2.0'))
        subprocess.run((DORADO_PATH[pfm], 'download', '--directory', model_dir,
            '--model', f'dna_r10.4.1_e8.2_400bps_{accuracy}@v4.2.0_5mCG_5hmCG@v2'))
    for modified_bases in '5mC', '6mA':
        subprocess.run((DORADO_PATH[pfm], 'download', '--directory', model_dir,
            '--model', f'dna_r10.4.1_e8.2_400bps_sup@v4.2.0_{modified_bases}@v2'))
    with open(dorado_config, 'w') as f:
        json.dump({'platform': pfm}, f)

def download_modkit(directory: str = MODKIT_DIR):
    """Download modkit

    Parameters
    ----------
    directory : str
        Destination directory for modkit
    """

    dest_tarfile = os.path.join(directory, 'modkit_v0.1.8_centos7_x86_64.tar.gz')
    print(f'downloading to {dest_tarfile}')
    if os.path.exists(dest_tarfile):
        raise RuntimeError(f'a file already exists at {dest_tarfile}')
    http = urllib3.PoolManager()
    with http.request('GET', MODKIT_URL, preload_content=False) as r, open(dest_tarfile, 'wb') as f:
        shutil.copyfileobj(r, f)
    with tarfile.open(dest_tarfile) as tar:
        tar.extractall(path=directory)


#===============================================================================
# intersect.py
#===============================================================================

"""intersect several bedmethyl files"""

# Imports ======================================================================

from pybedtools import BedTool
from functools import reduce




# Functions ====================================================================


def intersect_bedmethyl(bedmethyl, output_prefix, chromosomes=None):
    """Intersect two or more bedmethyl files

    Parameters
    ----------
    bedmethyl
        iterable of paths to bedmethyl formatted files of methylation results
    output_prefix
        prefix for output files
    chromosomes
        iterable of chromosomes to include in output
    """
    
    bedtools = tuple(BedTool(bed) for bed in bedmethyl)
    chromosomes=set(chromosomes) if chromosomes else None
    intersection = reduce(lambda a, b: a.intersect(b), bedtools)
    for bedtool, filename in zip((bt.intersect(intersection) for bt in bedtools), bedmethyl):
        bedtool.saveas(f'{output_prefix}-{filename}')



#===============================================================================
# check_gpu_availability.py
#===============================================================================

# Imports ======================================================================

import GPUtil
import nvsmi
import psutil
from tabulate import tabulate
from sys import exit
from time import sleep

def format_processes_table(processes):
    """Alternative string representation for a GPU process

    Parameters
    ----------
    processes
        an iterable of GPUprocess objects from nvsmi

    Returns
    -------
    str
        string representation of the processes  table
    """

    headers = ('GPU', 'PID', 'Process Name', 'GPU Memory Usage')
    table = ((p.gpu_id, p.pid, p.process_name,
              f'{int(p.used_memory)/1024:.2f} GB') for p in processes)

    return tabulate(table, headers=headers, tablefmt='rst')


def check_gpu_availability(devices, gpu_load: float = 0.9,
                           gpu_mem: float = 0.9):
    """Check for availability of indicated GPUs. If not available, prompt user
    to free up resources by terminating GPU processes. Terminate execution
    if GPUs are not made available.

    Parameters
    ----------
    devices
        iterable of devices ID's to check for
    gpu_load : float
        minimum GPU load availability (as a fraction of total)
    gpu_mem : float
        minimum GPU memory availability (as a fraction of total)
    """

    gpus=GPUtil.getGPUs()
    if not gpus:
        print('No GPUs found on this machine, terminating.')
        exit()
    available = GPUtil.getAvailable(limit=len(gpus), maxLoad=1-gpu_load,
                                    maxMemory=1-gpu_mem)
    unavailable = tuple(str(d) for d in devices if d not in available)
    if unavailable:
        processes = tuple(p for p in nvsmi.get_gpu_processes()
                    if p.gpu_id in unavailable)
        processes_table = format_processes_table(processes)
        reply = input('Cannot launch because the following processes are '
                    f'occupying resources on device(s) '
                    f'{", ".join(unavailable)}:\n\n{processes_table}\n\n'
                    f'Terminate these processes and continue? [y/N]:')
        if reply.casefold() in {'y', 'yes'}:
            for pid in {p.pid for p in processes}:
                psutil.Process(pid).terminate()
            sleep(4)
            check_gpu_availability(devices, gpu_load=gpu_load, gpu_mem=gpu_mem)
        else:
            print('GPU(s) not available, terminating.')
            exit()
    else:
        return True



#===============================================================================
# export_bedgraph.py
#===============================================================================

"""Export methylation data in bedgraph format"""

from pybedtools import BedTool

def export_bedgraph(bedmethyl: str, chromosomes=None, coverage: bool = False):
    """Convert methylation results to bedgraph format and print to stdout

    Parameters
    ----------
    bedmethyl : str
        path to a bedmethyl formatted file of methylation results
    chromosomes
        iterable of chromosomes to include in output
    coverage : bool
        if true, write a coverage track instead of a methylation track
    """

    bedtool = BedTool(bedmethyl).sort()
    if chromosomes:
        c = set(chromosomes)
        if coverage:
            for chrom, start, stop, *_, cov, _ in (i.fields for i in bedtool):
                if chrom in  c:
                    print(chrom, start, stop, cov, sep='\t')
        else:
            for chrom, start, stop, *_, methyl in (i.fields for i in bedtool):
                if chrom in  c:
                    print(chrom, start, stop, methyl, sep='\t')
    else:
        if coverage:
            for chrom, start, stop, *_, cov, _ in (i.fields for i in bedtool):
                print(chrom, start, stop, cov, sep='\t')
        else:
            for chrom, start, stop, *_, methyl in (i.fields for i in bedtool):
                print(chrom, start, stop, methyl, sep='\t')



import itertools
import os
import os.path
import subprocess
import pysam
import multiprocessing
from loreme.env import DORADO_PATH, DORADO_PATH_024, DORADO_MODEL_DIR, DORADO_PLATFORM

CPU_COUNT = multiprocessing.cpu_count()

def fast5_to_pod5(*input_fast5, output_pod5: str, threads: int = 1):
    """Convert FAST5 files to POD5 files

    Parameters
    ----------
    *input_fast5
        fast5 files or directories containing fast5s
    output_pod5 : str
        output POD5 file
    """

    parsed_fast5 = itertools.chain.from_iterable(
        ([os.path.join(f, f5) for f5 in os.listdir(f) if f5.endswith('.fast5')]
         if os.path.isdir(f) else [f]) for f in input_fast5)
    subprocess.run(('pod5', 'convert', 'fast5', *parsed_fast5,
                    '--output', output_pod5, '--threads', str(threads)))


def dorado_basecall(input_dir, output, speed: int = 400, accuracy: str ='fast',
           frequency: str = '4kHz', modified_bases: str = '5mCG_5hmCG',
           no_mod: bool = False, reference=None):
    """Run dorado basecaller

    Parameters
    ----------
    input_dir : str
        directory containing POD5 or Fast5 files
    output : str
        path to output BAM file
    speed : int
        pore speed, either 260 or 400 (default: 400)
    accuracy : str
        model accuracy, one of fast, hac, or sup (default: fast)
    frequency : str
        either 4kHz or 5kHz (default: 4kHz)
    modified_bases : str
        modified base model to use, one of 5mCG_5hmCG, 5mC, 6mA
        (default: 5mCG_5hmCG)
    reference : str
        path to reference index for alignment
    """

    if speed not in {260, 400}:
        raise RuntimeError('Invalid speed choice. Choose 400 or 260.')
    if accuracy not in {'fast', 'hac', 'sup'}:
        raise RuntimeError('Invalid accuracy choice. Choose fast, hac, or sup.')
    if frequency not in {'4kHz', '5kHz'}:
        raise RuntimeError('Invalid frequency choice. Choose 4kHz or 5kHz.')
    if modified_bases not in {'5mCG_5hmCG', '5mC', '6mA'}:
        raise RuntimeError('Invalid modified bases choice. Choose 5mCG_5hmCG, 5mC, 6mA.')
    if speed == 260 and frequency != '4kHz':
        raise RuntimeError('260bps speed only available at 4kHz frequency')
    if frequency == '4kHz' and modified_bases != '5mCG_5hmCG':
        raise RuntimeError('the only modified base model at 4kHz frequency is 5mCG_5hmCG')
    model = f"dna_r10.4.1_e8.2_{speed}bps_{accuracy}@v4.{1+(frequency=='5kHz')}.0"
    with open(output, 'wb') as f:
        subprocess.run((DORADO_PATH[DORADO_PLATFORM], 'basecaller',
                        os.path.join(DORADO_MODEL_DIR, model), input_dir)
                        + (not no_mod) * ('--modified-bases', modified_bases)
                        + bool(reference) * ('--reference', reference),
                       stdout=f)


def dorado_align(reference_index: str, input_reads: str, output_bam: str,
                 threads: int = CPU_COUNT, mem_per_thread_mb: int = 768):
    """Run dorado aligner, sorting and indexing the output

    Parameters
    ----------
    reference_index : str
        path to reference index
    input_reads : str
        path to input reads
    output_bam : str
        path to output BAM file
    """

    with open(output_bam, 'wb') as f, \
        subprocess.Popen((DORADO_PATH[DORADO_PLATFORM], 'aligner',
            reference_index, input_reads), stdout=subprocess.PIPE) as aligner:
        subprocess.run(('samtools', 'sort', '-@', str(threads),
                        '-m', f'{mem_per_thread_mb}M'),
                       stdin=aligner.stdout, stdout=f)
    pysam.index(output_bam)



import subprocess
from loreme.env import MODKIT_PATH

def modkit_pileup(fasta: str, input_bam: str, output_bed: str, log=None,
                  threads: int = 1):
    subprocess.run((MODKIT_PATH, 'pileup', input_bam, output_bed, '--ref', fasta,
                    '--threads', str(threads), '--preset', 'traditional')
                   + bool(log)*('--log-filepath', log))



#===============================================================================
# parse_gff.py
#===============================================================================

def generate_flank(genes_df, side: str = 'up', flank: int = 1000):
    """Generate coordinates of upstream or downstream flanks for input features

    Parameters
    ----------
    genes_df
        data frame containing gene coordinates
    side : str
        "up" for upstream flank, "down" for downstream flank
    flank : int
        size of flank in bp

    Yields
    ------
    tuple
        coordinates, strand, and gene ID corresponding to the flank
    """

    if side == 'up':
        for index, row in genes_df.iterrows():
            if row['strand'] == '+':
                yield row['seqid'], int(row['start'] - flank), row['start'], '+', index
            elif row['strand'] == '-':
                yield row['seqid'], row['end'], int(row['end'] + flank), '-', index
    elif side == 'down':
        for index, row in genes_df.iterrows():
            if row['strand'] == '+':
                yield row['seqid'], row['end'], int(row['end'] + flank), '+', index
            elif row['strand'] == '-':
                yield row['seqid'], int(row['start'] - flank), row['start'], '-', index


def generate_feature_body(features_df, upstream_flank: int = 0, downstream_flank: int = 0):
    """Generate coordinates for bodies of input features

    Parameters
    ----------
    features_df
        data frame containing gene coordinates
    upstream_flank : int
        size of upstream region to include in body, in bp
    downstream_flank : int
        size of downstream region to include in body, in bp

    Yields
    ------
    tuple
        coordinates, strand, and feature ID corresponding to the feature body
    """

    for index, row in features_df.iterrows():
        if row['strand'] == '+':
            yield row['seqid'], int(row['start'] - upstream_flank) , int(row['end'] + downstream_flank), row['strand'], index
        if row['strand'] == '-':
            yield row['seqid'], int(row['start'] - downstream_flank) , int(row['end'] + upstream_flank), row['strand'], index


def generate_promoter(genes_df, upstream_flank: int = 2000, downstream_flank: int = 0):
    """Generate coordinates for promoters of input genes

    Parameters
    ----------
    genes_df
        data frame containing gene coordinates
    upstream_flank : int
        size of upstream region to include in promoter, in bp
    downstream_flank : int
        size of downstream region to include in promoter, in bp

    Yields
    ------
    tuple
        coordinates, strand, and feature ID corresponding to the promoter
    """

    for index, row in genes_df.iterrows():
        if row['strand'] == '+':
            yield row['seqid'], int(row['start'] - upstream_flank), int(row['start'] + downstream_flank), '+', index
        elif row['strand'] == '-':
            yield row['seqid'], int(row['end'] - downstream_flank), int(row['end'] + upstream_flank), '-', index



#===============================================================================
# gene_body_methylation.py
#===============================================================================

"""Quantify gene body methylation"""

# Imports ======================================================================

from pybedtools import BedTool
import pandas as pd
import gff2bed

from loreme.parse_gff import generate_feature_body
from loreme.plot_bedtools import generate_feature_data
from loreme.methylation_hist import methylation_hist




# Functions ====================================================================

def gene_body_methylation(features, bedmethyl, upstream_flank: int = 0,
                         downstream_flank: int = 0, chromosomes=None,
                         cytosines: bool = False, coverage: bool = False,
                         min_coverage: bool = 1, bins: bool = False, hist=None,
                         levels=['Min', 'Low', 'Mid', 'High', 'Max'],
                         palette: str = 'mako_r', hist_log=False):
    """Quantify methylation in gene bodies

    Parameters
    ----------
    features
        path to GFF3 file containing gene coordinates
    bedmethyl
        path to bedmethyl formatted file containing methylation results
    upstream_flank : int
        size of upstream flank to include in gene body, in bp [0]
    downstream_flank : int
        size of downstream flank to include in gene body, in bp [0]
    chromosomes
        iterable of chromosomes to include, or None to include all chromosomes
    coverage : bool
        if True, include a coverage column in the results
    min_coverage : int
        minimum coverage for a gene to be included
    bins : bool
        if True, add an extra column binning genes by methylation level [False]
    hist
        if given, write a histogram of gene bin methylation levels
    hist_log
        if true, use a log scale for the histogram
    levels
        iterable of labels for binning by methylation level
    palette : str
        color palette for histogram
    """

    genes = pd.DataFrame((f for f in gff2bed.parse(features) if f[1] > upstream_flank),
        columns=('seqid', 'start', 'end', 'strand', 'attributes'))
    genes.index = (attr['ID'] for attr in genes['attributes'])
    gene_body = BedTool(tuple(generate_feature_body(genes,
        upstream_flank=upstream_flank, downstream_flank=downstream_flank)))
    methyl = BedTool(bedmethyl)
    methyl_gene_body = methyl.intersect(gene_body, wo=True)
    df = pd.DataFrame(
        generate_feature_data(methyl_gene_body, chromosomes=chromosomes),
        columns=('chrom', 'start', 'end', 'gene', 'strand', 'cytosines', 'coverage', 'methyl_sum')
    ).groupby(by=['chrom', 'start', 'end', 'gene', 'strand'], as_index=False).sum().sort_values(by=['chrom', 'start'])
    df.index = df['gene']
    df['Methylation level (%)'] = df['methyl_sum'] / df['cytosines']
    df['Discrete level'] = pd.qcut(df['Methylation level (%)'].rank(method='first'), q=len(levels), labels=levels)
    if hist:
        methylation_hist(df, hist, palette=palette, log_scale=hist_log)
    for _, (chrom, start, end, gene, strand, cyt, cov, methyl, level) in df.iloc[:,[0,1,2,3,4,5,6,8,9]].iterrows():
        row = (chrom, start, end, gene, f'{methyl:.2f}', strand) + cytosines*(cyt,) + coverage*(cov,) + bins*(level,)
        if cov >= min_coverage:
            print(*row, sep='\t')
